; BOOTLOADER FOR LBA DISKS, SUPPORTS FAT16
; CAN BOOT FROM PARTITIONED AND UNPARTITIONED DISKS

[org 0xFc24]		; code offset

;  This is a special bootloader for USB sticks

; initialize data segment and stack
xor ax, ax
mov ds, ax
mov es, ax

mov byte [0x7C15], dl ; save the boot drive

; move to the end of the segment
mov si, 0x7c00
mov di, 0xFc00
mov cx, 0x100
cld
rep movsw

;mov bx, start
;jmp bx
db 0xE9
dw 0x8000

start:
mov ss, ax
retry:
mov sp, 0x400

; read the first sector of root directory into 0000:0600
mov cl, [0xFC10] ; FAT copies
xor ax, ax
fat_loop1:
add ax, [0xFC16] ; sectors per FAT
loop fat_loop1

; number of root directory entries
mov cx, [0xFc11]
read_root_sector:
mov bx, 0x600
mov si, bx
xor dx, dx
call read_sector
root_loop:
;mov byte [si+11], 0
;call puts ; print file name

call kernelfilecmp

jz found_kernel

dec cx
jz nokernelfound

add si, 32
add bx, 32
cmp bx, 0x800
je next_root_sector
jmp root_loop
next_root_sector:
inc ax
jmp read_root_sector

found_kernel:
; bx == pointer to directory entry of the kernel file
; bx + 28 == file size
; bx + 26 == starting cluster
;mov si, kernelfound
;call puts

xor ch, ch

mov ax, [bx+26] ; ax = starting cluster of the kernel file

; load kernel to 0x600

mov bx, 0x600
read_kernel_loop:
call read_cluster

call next_cluster

cmp ax, 0xFFF7
jb read_kernel_loop


; follow kernel file's cluster chain and load the file to memory...

; jump to kernel
jmp 0x60:0


nokernelfound:
mov si, nokernelmsg
call puts
halt:
jmp halt

; ax == cluster
; returns: cluster value to ax
next_cluster:
  push bx
  push cx

; offset to the cluster: cluster * 2
  xor dl, dl
  add ax, ax
; store the more significant bits of the offset to dl
  rcr dl, 1

; bx = offset to the cluster

  mov bx, ax
  ; bx = bx % 512
  and bx, 0x01FF

; save the offset to stack
  push bx

; divide with 512 (sector size)
  mov cl, 9
  shr ax, cl
; and keep the more significant bit while dividing
  or al, dl
  xor dl, dl

  mov cl, 1
  cmp bx, 0x1FF
  jne read_one_sector
  inc cl
  read_one_sector:

  ; --- <caching> ---
  cmp ax, [0xFFFE] ; are the required FAT sectors already in memory?
  jne not_in_cache
  cmp dl, [0xFC01]
  jne not_in_cache
  cmp cl, [0xFC00]
  je read_cluster_table_end

  not_in_cache:
  mov [0xFFFE], ax
  mov [0xFC00], cl
  mov [0xFC01], dl
  ; --- </caching> ---

; read one or two sectors from sector ax
  mov bx, 0xF800
  read_cluster_table:
  call read_sector
  dec cl
  jz read_cluster_table_end
  add bh, 0x02
  add ax, 1
  adc dx, 0
; read also the next sector
  jmp read_cluster_table
  read_cluster_table_end:
  pop bx

  ; bx += 0xF800
  add bx, 0xF800

  mov ax, [bx]

  pop cx
  pop bx
ret

; returns 0 if the string pointed by SI register is identical with kernel file name
kernelfilecmp:
  push si
  push cx

  mov di, kernelfile
  mov cx, 11
  repe cmpsb
  
  pop cx
  pop si
ret

; reads cluster
; ax == cluster to read
; es:bx == pointer to buffer
; increments bx by the amount of read bytes
read_cluster:
  push ax
  push cx

  ; sector to read = (cluster-2) * sectors per cluster
  xor cx, cx
  mov dx, cx
  mov cl, [0xFC0D]
  sub ax, 2
  mul cx

  mov cl, [0xFC10]
  read_cluster_asloop:
  add ax, [0xFC16]
  adc dl, 0
  dec cl
  jz read_cluster_asloop_end
  jmp read_cluster_asloop
  read_cluster_asloop_end:
  mov di, [0xFC11]
  mov cl, 4
  shr di, cl
  add ax, di
  adc dl, 0

  mov cl, [0xFC0D]
  read_cluster_loop:
  call read_sector
  add bh, 0x02
  dec cl
  jz read_cluster_end
  add ax, 1
  adc dx, 0
  jmp read_cluster_loop

  read_cluster_end:
  pop cx
  pop ax
ret

lba_packet:
db 0x10 ; size of packet
db 0x00 ; reserved
dw 0 ; number of blocks to transfer
dw 0 ; transfer buffer (offset)
dw 0 ; transfer buffer (segment)
dw 0 ; block number (bits 0-15)
dw 0 ; block number (bits 16-31)
dw 0 ; block number (bits 32-47)
dw 0 ; block number (bits 48-63)

; reads logical sector from 1,44 MB floppy
; dx:ax == sector
; es:bx = pointer to buffer
read_sector:
  push ax
  push cx
  push dx
  push si
  mov cx, 16 ; number of retries
  
  add ax, [0xFC1C] ; hidden sectors
  adc dx, [0xFC1E]
  add ax, [0xFC0E] ; reserved sectors
  adc dx, 0
    

  mov si, lba_packet
  mov word [si+4], bx
  mov word [si+6], es
  mov word [si+8], ax
  mov word [si+10], dx

  mov dl, byte [0xFC15]


;  write dot
;  push ax
  push bx

  mov bx, 0x0007
  mov ax, 0x0E2E
  int 0x10

  pop bx
;  pop ax


  read_sector_retry:
  mov word [si+2], 0x0001
  mov ah, 0x42
  int 0x13
  jnc read_sector_end

  xor ah, ah
  int 0x13

  dec cx
  jnz read_sector_retry

  mov si, error
  call puts
  jmp retry

  read_sector_end:

  pop si
  pop dx
  pop cx
  pop ax
ret

; si == pointer to string
puts:
;  push bx
;  push ax
;  push si

  mov bx, 0x0007
  mov ah, 0x0E

  puts_loop:
  mov al, [si]
  test al, 0xFF
  jz puts_end
  int 0x10
  inc si
  jmp puts_loop

  puts_end:

  mov al, 0x0D
  int 0x10
  mov al, 0x0A
  int 0x10

;  pop si
;  pop ax
;  pop bx
ret

kernelfile: db "KERNEL  BIN"
nokernelmsg: db "NOKRN "
error: db "ERR", 0

times 0x01FE-0x24-($-$$) db 0

dw 0xAA55
